import util;
import std.math;
import std.file;
import std.string;
import SDL;

int i;
char[] rna;
int verbose;

struct Color {
    int r, g, b, a;
    static Color opCall(int r, int g, int b, int a) {
        Color c;
        c.r = r;
        c.g = g;
        c.b = b;
        c.a = a;
        return c;
    }

    bool opEqual(Color c) {
        return r == c.r && g == c.g && b == c.b && a == c.a;
    }

    void print() {
        printf("r=%d g=%d b=%d a=%d\n", r, g, b, a);
    }
}
Color black, red, green, yellow, blue, magenta, cyan, white;
Color transparent, opaque;

Color[] bucket;

enum Dir {
    N, E, W, S,
};
Dir dir;
alias int[2] Pos;
Pos pos, mark;

alias Color[600][600] Bitmap;
Bitmap[] bitmap;
Color cache_color;
bool cache_avail;

Color current_pixel() {
    if (cache_avail) {
        return cache_color;
    }

    int rgb_num = 0, alpha_num = 0;
    int r = 0, g = 0, b = 0, a = 0;
    //for (int i = 0; i < bucket.length; i++) {
    foreach (Color c; bucket) {
        //Color c = bucket[i];
        if (c.a >= 0) {
            alpha_num++;
            a += c.a;
        }
        else {
            rgb_num++;
            r += c.r;
            g += c.g;
            b += c.b;
        }
    }
    //printf("%d %d %d %d %d %d %d\n", bucket.length, r,g,b,a,rgb_num,alpha_num);
    if (rgb_num) {
        r = r / rgb_num;
        g = g / rgb_num;
        b = b / rgb_num;
    }
    else {
        r = g = b = 0;
    }
    if (alpha_num) {
        a = a / alpha_num;
    }
    else {
        a = 255;
    }
    cache_avail = true;
    cache_color = Color(r * a / 255, g * a / 255, b * a / 255, a);
    return cache_color;
}

Color get_pixel(int x, int y) {
    return bitmap[$-1][y][x];
}

void set_pixel(int x, int y) {
    //printf("%d %d\n", x, y);
    bitmap[$-1][y][x] = current_pixel();
}

void line() {
    int dx = mark[0] - pos[0];
    int dy = mark[1] - pos[1];
    int d = max(abs(dx), abs(dy));
    int c = dx * dy <= 0 ? 1 : 0;
    int x = pos[0] * d + (d - c) / 2;
    int y = pos[1] * d + (d - c) / 2;
    if (verbose) printf("line (%d,%d) - (%d,%d)\n", mark[0], mark[1], pos[0], pos[1]);
    for (int j = 0; j < d; j++) {
        //printf("line %d %d %d\n", j, x, y);
        set_pixel(x / d, y / d);
        x += dx;
        y += dy;
    }
    set_pixel(mark[0], mark[1]);
}

void move() {
    switch (dir) {
    case Dir.N: {
        pos[1] = pos[1] == 0 ? 599 : pos[1] - 1;
        break;
    }
    case Dir.E: {
        pos[0] = pos[0] == 599 ? 0 : pos[0] + 1;
        break;
    }
    case Dir.W: {
        pos[0] = pos[0] == 0 ? 599 : pos[0] - 1;
        break;
    }
    case Dir.S: {
        pos[1] = pos[1] == 599 ? 0 : pos[1] + 1;
        break;
    }
    }
}

void turnCCW() {
    switch (dir) {
    case Dir.N: {
        dir = Dir.W;
        break;
    }
    case Dir.E: {
        dir = Dir.N;
        break;
    }
    case Dir.W: {
        dir = Dir.S;
        break;
    }
    case Dir.S: {
        dir = Dir.E;
        break;
    }
    }
}

void turnCW() {
    switch (dir) {
    case Dir.N: {
        dir = Dir.E;
        break;
    }
    case Dir.E: {
        dir = Dir.S;
        break;
    }
    case Dir.W: {
        dir = Dir.N;
        break;
    }
    case Dir.S: {
        dir = Dir.W;
        break;
    }
    }
}

void fill(int ix, int iy, Color c) {
    int[] st;
    st ~= (ix << 12) | iy;
    while (st.length) {
        int s = pop(st);
        int x = s >> 12;
        int y = s & 4095;
        if (get_pixel(x, y) != c) continue;
        set_pixel(x, y);
        if (x > 0) {
            st ~= ((x-1)<<12)|y;
        }
        if (x < 599) {
            st ~= ((x+1)<<12)|y;
        }
        if (y > 0) {
            st ~= (x<<12)|(y-1);
        }
        if (y < 599) {
            st ~= (x<<12)|(y+1);
        }
    }
}

void tryfill() {
    if (verbose) printf("fill (%d,%d)\n", pos[0], pos[1]);
    Color n = current_pixel();
    Color o = get_pixel(pos[0], pos[1]);
    if (n != o) fill(pos[0], pos[1], o);
}

void add_bitmap(int alpha = 0) {
    if (verbose) printf("add bitmap\n");
    if (bitmap.length >= 10) return;
    bitmap.length = bitmap.length + 1;
    for (int y = 0; y < 600; y++) {
        for (int x = 0; x < 600; x++) {
            bitmap[$-1][y][x].r = 0;
            bitmap[$-1][y][x].g = 0;
            bitmap[$-1][y][x].b = 0;
            bitmap[$-1][y][x].a = alpha;
        }
    }
}

void compose() {
    if (verbose) printf("compose\n");
    if (bitmap.length < 2) return;
    for (int y = 0; y < 600; y++) {
        for (int x = 0; x < 600; x++) {
            Color c0 = bitmap[$-1][y][x];
            Color c1 = bitmap[$-2][y][x];
            bitmap[$-2][y][x].r = c0.r + c1.r * (255 - c0.a) / 255;
            bitmap[$-2][y][x].g = c0.g + c1.g * (255 - c0.a) / 255;
            bitmap[$-2][y][x].b = c0.b + c1.b * (255 - c0.a) / 255;
            bitmap[$-2][y][x].a = c0.a + c1.a * (255 - c0.a) / 255;
        }
    }
    bitmap.length = bitmap.length - 1;
}

void clip() {
    if (verbose) printf("clip\n");
    if (bitmap.length < 2) return;
    for (int y = 0; y < 600; y++) {
        for (int x = 0; x < 600; x++) {
            Color c0 = bitmap[$-1][y][x];
            Color c1 = bitmap[$-2][y][x];
            bitmap[$-2][y][x].r = c1.r * c0.a / 255;
            bitmap[$-2][y][x].g = c1.g * c0.a / 255;
            bitmap[$-2][y][x].b = c1.b * c0.a / 255;
            bitmap[$-2][y][x].a = c1.a * c0.a / 255;
        }
    }
    bitmap.length = bitmap.length - 1;
}

void rna_build() {
    int inst_cnt = 0;
    int cmd_cnt = 0;
    while (i+6 < rna.length) {
        if (verbose) printf("%d(%d) ", inst_cnt++, cmd_cnt++);
        //if (inst_cnt > 82000) break;
        switch (rna[i..i+7]) {
        case "PIPIIIC": {
            if (verbose) printf("add black\n");
            cache_avail = false;
            bucket ~= black;
            break;
        }
        case "PIPIIIP": {
            if (verbose) printf("add red\n");
            cache_avail = false;
            bucket ~= red;
            break;
        }
        case "PIPIICC": {
            if (verbose) printf("add green\n");
            cache_avail = false;
            bucket ~= green;
            break;
        }
        case "PIPIICF": {
            if (verbose) printf("add yellow\n");
            cache_avail = false;
            bucket ~= yellow;
            break;
        }
        case "PIPIICP": {
            if (verbose) printf("add blue\n");
            cache_avail = false;
            bucket ~= blue;
            break;
        }
        case "PIPIIFC": {
            if (verbose) printf("add magenta\n");
            cache_avail = false;
            bucket ~= magenta;
            break;
        }
        case "PIPIIFF": {
            if (verbose) printf("add cyan\n");
            cache_avail = false;
            bucket ~= cyan;
            break;
        }
        case "PIPIIPC": {
            if (verbose) printf("add white\n");
            cache_avail = false;
            bucket ~= white;
            break;
        }
        case "PIPIIPF": {
            if (verbose) printf("add trans\n");
            cache_avail = false;
            bucket ~= transparent;
            break;
        }
        case "PIPIIPP": {
            if (verbose) printf("add opaque\n");
            cache_avail = false;
            bucket ~= opaque;
            break;
        }
        case "PIIPICP": {
            if (verbose) printf("clear bucket\n");
            cache_avail = false;
            bucket.length = 0;
            break;
        }
        case "PIIIIIP": {
            move();
            break;
        }
        case "PCCCCCP": {
            turnCCW();
            break;
        }
        case "PFFFFFP": {
            turnCW();
            break;
        }
        case "PCCIFFP": {
            mark[] = pos[];
            break;
        }
        case "PFFICCP": {
            line();
            break;
        }
        case "PIIPIIP": {
            tryfill();
            break;
        }
        case "PCCPFFP": {
            add_bitmap();
            break;
        }
        case "PFFPCCP": {
            compose();
            break;
        }
        case "PFFICCF": {
            clip();
            break;
        }
        default: {
            inst_cnt--;
            //printf("do nothing\n");
        }
        }
        i += 7;
    }
}

void set_rgb(Color* c, int r, int g, int b) {
    c.r = r;
    c.g = g;
    c.b = b;
    c.a = -1;
}

void set_alpha(Color* c, int a) {
    c.r = 0;
    c.g = 0;
    c.b = 0;
    c.a = a;
}

void init() {
    set_rgb(&black, 0,0,0);
    set_rgb(&red, 255,0,0);
    set_rgb(&green, 0,255,0);
    set_rgb(&yellow, 255,255,0);
    set_rgb(&blue, 0,0,255);
    set_rgb(&magenta, 255,0,255);
    set_rgb(&cyan, 0,255,255);
    set_rgb(&white, 255,255,255);
    set_alpha(&transparent, 0);
    set_alpha(&opaque, 255);

    dir = Dir.E;
    pos[0] = 0;
    pos[1] = 0;

    add_bitmap(255);
    //bitmap.length = 1;
}

void dumpbmp(Bitmap bitmap, char[] filename) {
    SDL_Init(SDL_INIT_VIDEO);
    SDL_Surface* s = SDL_CreateRGBSurface(SDL_SWSURFACE, 600, 600, 32, 0,0,0,0);
    Uint8* p = cast(Uint8*)s.pixels;
    for (int y = 0; y < 600; y++) {
        for (int x = 0; x < 600; x++) {
            Color c = bitmap[y][x];
            Uint32* o = cast(Uint32*)(p + y * s.pitch + x * 4);
            *o = SDL_MapRGB(s.format, c.r, c.g, c.b);
        }
    }
    SDL_SaveBMP(s, toStringz(filename));
}

int main(char[][] args) {
    init();

    if (args.length > 1 && args[1] == "-v") {
        verbose = 1;
        args = args[1..$];
    }

    if (args.length == 2) {
        printf("use %.*s\n", args[1]);
        rna ~= cast(char[])std.file.read(args[1]);
    }
    else {
        rna ~= cast(char[])std.file.read(`endo.rna`);
    }

    rna_build();

    if (bitmap.length > 1) {
        printf("owattenai yo!\n");
        for (int i = 0; i < bitmap.length; i++) {
            dumpbmp(bitmap[i], std.string.toString(i)~".bmp");
        }
        while (bitmap.length > 1) {
            compose();
        }
    }

    /*
    add_bitmap();
    //compose();
    bucket = null;
    bucket ~= opaque;
    tryfill();
    clip();
    */

    dumpbmp(bitmap[$-1], "endo.bmp");

    return 0;
}
